/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.createDataDictionary;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFacade.addUserHeaderInclude;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFactory.createCSourcePackage;
import static org.fortiss.af3.generator.common.utils.SourceModelElementFactory.createByteContentUnitForPluginFile;
import static org.fortiss.af3.platform.raspberry.generator.executable.HeaderCopyGenerator.copyHeaderFiles;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getConfigureFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getEclipseProjectFile;
import static org.fortiss.af3.platform.raspberry.generator.templates.RasPiCTemplates.getMakedefsFile;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getRootElement;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.TransformationUtils.createTransformedObjectFor;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.List;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.generator.ComponentProgram;
import org.fortiss.af3.expression.model.DataDictionary;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.generator.common.model.c.AF3GeneratorCommonLanguagesCFactory;
import org.fortiss.af3.generator.common.model.c.CSourcePackage;
import org.fortiss.af3.generator.common.model.source.AbstractUnit;
import org.fortiss.af3.generator.common.model.source.ByteContentUnit;
import org.fortiss.af3.generator.common.model.source.SourceUnit;
import org.fortiss.af3.platform.language.executable.ExecutionUnitExecutableBase;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.raspberry.AF3PlatformRaspberryActivator;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;
import org.fortiss.tooling.kernel.extension.exception.ChainTransformationFailedException;

/**
 * {@link ExecutionUnitExecutableBase} for the {@link RaspberryPi} platform.
 * 
 * @author eder
 * @author hoelzl
 * @author $Author: eder $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class RaspberryPIExecutable extends ExecutionUnitExecutableBase<RaspberryPi, CSourcePackage> {
	/** Folder name for library source files. */
	private static final String SRC_LIB_SUB_PACKAGE_NAME = "src-lib";
	/** Folder name for library include files. */
	public static final String INC_LIB_SUB_PACKAGE_NAME = "inc-lib";
	/** Folder name for linkable static library files. */
	private static final String LIB_SUB_PACKAGE_NAME = "lib";
	/** Default library names to be used always. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES = "af3pihal rt pthread";
	/** Default library names extended with brick. */
	// Note that the order is important (GCC linker crap!)
	private static final String LIB_NAMES_WITH_BRICK = "brick " + LIB_NAMES;

	/** Constructor. */
	public RaspberryPIExecutable(RaspberryPi modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	protected CSourcePackage createExecutionUnitSourcePackage(String name,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		// create the result package
		CSourcePackage sourcePackage = createCSourcePackage();
		sourcePackage.setBaseLocation(name);
		try {
			addDataDictionaryCode(context, sourcePackage);
			addLogicalComponentCode(deployedComponents, context, sourcePackage);
			addEclipseCProjectFiles(sourcePackage, modelElement.getName());
			addStaticLibraries(sourcePackage);
			addConfigureAndMakedefsFiles(sourcePackage, modelElement.getName(),
					LIB_NAMES_WITH_BRICK);
			addMainFile(sourcePackage, deployedComponents, deployedPorts, context);
		} catch(Exception ex) {
			error(AF3PlatformRaspberryActivator.getDefault(), ex.getMessage(), ex);
			ex.printStackTrace();
		}
		return sourcePackage;
	}

	/** Creates the main file by using the {@link MainGenerator} helper class. */
	private void addMainFile(CSourcePackage sourcePackage,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {
		CSourcePackage srcGenPack = (CSourcePackage)sourcePackage.getSrcGenPackage();
		MainGenerator mg =
				new MainGenerator(modelElement, deployedComponents, deployedPorts, context);
		srcGenPack.addUnit(mg.createSingleUnitMain());
	}

	/** Adds the build process files: configure and Makedefs. */
	private void addConfigureAndMakedefsFiles(CSourcePackage sourcePackage, String binaryName,
			String libNames) {
		sourcePackage.addUnit(getConfigureFile(binaryName, libNames));
		sourcePackage.addUnit(getMakedefsFile());
	}

	/** Adds the library code need by the deployment code generator. */
	private void addStaticLibraries(CSourcePackage sourcePackage) throws Exception {
		CSourcePackage srcLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		srcLib.setBaseLocation(SRC_LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(srcLib);

		CSourcePackage incLib =
				AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		incLib.setBaseLocation(INC_LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(incLib);

		CSourcePackage lib = AF3GeneratorCommonLanguagesCFactory.eINSTANCE.createCSourcePackage();
		lib.setBaseLocation(LIB_SUB_PACKAGE_NAME);
		sourcePackage.getSubPackages().add(lib);
		lib.addUnit(getStaticLibraryUnit("libbrick.a"));
		lib.addUnit(getStaticLibraryUnit("libaf3pihal.a"));
		copyHeaderFiles(incLib);
	}

	/** Creates a {@link ByteContentUnit} for the given static library. */
	private AbstractUnit getStaticLibraryUnit(String libName) throws Exception {
		return createByteContentUnitForPluginFile(AF3PlatformRaspberryActivator.PLUGIN_ID,
				"code-gen-hal/lib", libName, false);
	}

	/** Adds auxiliary files for Eclipse C project. */
	private void addEclipseCProjectFiles(CSourcePackage sourcePackage, String projectName) {
		sourcePackage.addUnit(getEclipseProjectFile(projectName));
	}

	/** Creates the code for all deployed components. */
	private void addLogicalComponentCode(List<Pair<ExecutionUnit, Component>> deployedComponents,
			ITransformationContext context, CSourcePackage sourcePackage)
			throws ChainTransformationFailedException {
		for(Pair<ExecutionUnit, Component> pair : deployedComponents) {
			if(pair.getFirst() != modelElement) {
				continue;
			}
			Component cmp = pair.getSecond();
			ComponentProgram cprog =
					createTransformedObjectFor(cmp, ComponentProgram.class, context);
			CSourcePackage cPack = createTransformedObjectFor(cprog, CSourcePackage.class, context);
			cPack.mergeInto(sourcePackage);
		}
	}

	/** Creates the code for the data dictionary. */
	private void
			addDataDictionaryCode(ITransformationContext context, CSourcePackage sourcePackage)
					throws ChainTransformationFailedException {
		DataDictionary dd = getRootElement(modelElement, DataDictionary.class);
		if(dd == null) {
			dd = createDataDictionary();
		}
		CSourcePackage ddPackage = createTransformedObjectFor(dd, CSourcePackage.class, context);
		ddPackage.mergeInto(sourcePackage);
	}

	private static void addFilesWithReference(CSourcePackage pkg, SourceUnit source,
			String fileWithoutExtension) throws IOException, URISyntaxException {
		addFile(pkg, "src/" + fileWithoutExtension + ".c");
		addFile(pkg, "inc/" + fileWithoutExtension + ".h");
		addSystemCReference(source, fileWithoutExtension);
	}

	private static void addFile(CSourcePackage pkg, String file) throws IOException,
			URISyntaxException {
		pkg.addUnit(createByteContentUnitForPluginFile(AF3PlatformRaspberryActivator.PLUGIN_ID,
				"lib", file, false));
	}

	private static void addSystemCReference(SourceUnit source, String reference) {
		addUserHeaderInclude(source, reference);
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getInitialization() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getTermination() {
		return null;
	}
}
